{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module Loop where {

import qualified List;
import qualified Term;
import qualified HOL;
import qualified Nat;
import qualified Product_Type;
import qualified Trs;
import qualified Position;

rewrites ::
  forall a b.
    (Eq a,
      Eq b) => (Term.Term a b, Term.Term a b) ->
                 Term.Ctxt a b ->
                   Term.Sub a b ->
                     (Term.Term a b, Term.Term a b) ->
                       [(Term.Term a b, Term.Term a b)] -> Bool;
rewrites (s, t) c sigma (l, r) ra =
  HOL.eqop s (Term.ctxt_apply c (Term.sub_apply l sigma)) &&
    (HOL.eqop t (Term.ctxt_apply c (Term.sub_apply r sigma)) &&
      List.member (l, r) ra);

subst_safe ::
  forall a b.
    (Eq a,
      Eq b) => [a] ->
                 [(Term.Term b a, Term.Term b a)] ->
                   Term.Sub b a -> Maybe (Term.Sub b a);
subst_safe vs ((Term.Var x, t) : ts) sigma =
  (if List.member (Term.Var x, t) ts then Loop.subst_safe vs ts sigma
    else (if List.member x vs &&
               not (HOL.eqop (Term.sub_apply (Term.Var x) sigma) t)
           then Nothing
           else (case Loop.subst_safe (x : vs) ts sigma of {
                  Nothing -> Nothing;
                  Just a ->
                    let {
                      (Term.Sub s) = a;
                    } in Just (Term.Sub
                                (\ y -> (if HOL.eqop x y then t else s y)));
                })));
subst_safe vs [] sigma = Just sigma;

decomp :: forall a b. [a] -> [b] -> Maybe [(a, b)];
decomp (s : ss) (t : ts) =
  (case Loop.decomp ss ts of {
    Nothing -> Nothing;
    Just pairs -> Just ((s, t) : pairs);
  });
decomp [] (uw : ux) = Nothing;
decomp (uu : uv) [] = Nothing;
decomp [] [] = Just [];

decompose_one ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_one (Term.Fun f ts) (Term.Fun g ss) =
  (if HOL.eqop f g then Loop.decomp ts ss else Nothing);
decompose_one (Term.Fun v va) (Term.Var y) = Nothing;
decompose_one (Term.Var x) t = Just [(Term.Var x, t)];

decompose_n ::
  forall a b.
    (Eq a) => Nat.Nat ->
                Term.Term a b ->
                  Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_n (Nat.Suc n) s t =
  (case Loop.decompose_n n s t of {
    Nothing -> Nothing;
    Just a ->
      List.foldla
        (\ aa (b @ (sa, ta)) ->
          (case aa of {
            Nothing -> Nothing;
            Just ys ->
              (case Loop.decompose_one sa ta of {
                Nothing -> Nothing;
                Just xs -> Just (List.append xs ys);
              });
          }))
        (Just []) a;
  });
decompose_n Nat.Zero_nat s t = Loop.decompose_one s t;

decompose ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose s t = Loop.decompose_n (Term.sz s) s t;

decompose_list ::
  forall a b.
    (Eq a) => [(Term.Term a b, Term.Term a b)] ->
                Maybe [(Term.Term a b, Term.Term a b)];
decompose_list ts =
  List.foldla
    (\ a (b @ (s, t)) ->
      (case a of {
        Nothing -> Nothing;
        Just ls ->
          (case Loop.decompose s t of {
            Nothing -> Nothing;
            Just ys -> Just (List.append ls ys);
          });
      }))
    (Just []) ts;

match_all ::
  forall a b.
    (Eq a, Eq b) => [(Term.Term a b, Term.Term a b)] -> Maybe (Term.Sub a b);
match_all ts =
  (case Loop.decompose_list ts of {
    Nothing -> Nothing;
    Just ls -> Loop.subst_safe [] ls (Term.Sub Term.Var);
  });

getpr ::
  forall a b.
    (Eq a,
      Eq b) => Term.Term a b ->
                 Term.Term a b ->
                   Position.Pos ->
                     (Term.Term a b, Term.Term a b) ->
                       Maybe (Term.Term a b,
                               (Term.Ctxt a b,
                                 (Term.Sub a b,
                                   (Term.Term a b, Term.Term a b))));
getpr s t p (l, r) =
  (case Loop.match_all [(l, Term.subt_at s p), (r, Term.subt_at t p)] of {
    Nothing -> Nothing;
    Just sigma ->
      (if Loop.rewrites (s, t) (Trs.ctxt_of_pos_term p s) sigma (l, r) [(l, r)]
        then Just (s, (Trs.ctxt_of_pos_term p s, (sigma, (l, r))))
        else Nothing);
  });

getp ::
  forall a b.
    (Eq a,
      Eq b) => Term.Term a b ->
                 Term.Term a b ->
                   Position.Pos ->
                     [(Term.Term a b, Term.Term a b)] ->
                       Maybe (Term.Term a b,
                               (Term.Ctxt a b,
                                 (Term.Sub a b,
                                   (Term.Term a b, Term.Term a b))));
getp s t p (r : rs) =
  (case Loop.getpr s t p r of {
    Nothing -> Loop.getp s t p rs;
    Just a -> Just a;
  });
getp s t p [] = Nothing;

find_get_p ::
  forall f v.
    (Eq f,
      Eq v) => Term.Term f v ->
                 Term.Term f v ->
                   [Position.Pos] ->
                     [(Term.Term f v, Term.Term f v)] ->
                       Maybe (Term.Term f v,
                               (Term.Ctxt f v,
                                 (Term.Sub f v,
                                   (Term.Term f v, Term.Term f v))));
find_get_p s t (p : ps) r =
  (case Loop.getp s t p r of {
    Nothing -> Loop.find_get_p s t ps r;
    Just a -> Just a;
  });
find_get_p s t [] r = Nothing;

get ::
  forall a b.
    (Eq a,
      Eq b) => Term.Term a b ->
                 Term.Term a b ->
                   [(Term.Term a b, Term.Term a b)] ->
                     Maybe (Term.Term a b,
                             (Term.Ctxt a b,
                               (Term.Sub a b, (Term.Term a b, Term.Term a b))));
get s t r = Loop.find_get_p s t (Term.poss_list s) r;

get_list ::
  forall a b.
    (Eq a,
      Eq b) => [(Term.Term a b)] ->
                 Term.Term a b ->
                   [(Term.Term a b, Term.Term a b)] ->
                     Maybe [(Term.Term a b,
                              (Term.Ctxt a b,
                                (Term.Sub a b,
                                  (Term.Term a b, Term.Term a b))))];
get_list (s : t : ts) u r =
  (case Loop.get s t r of {
    Nothing -> Nothing;
    Just d ->
      (case Loop.get_list (t : ts) u r of {
        Nothing -> Nothing;
        Just ls -> Just (d : ls);
      });
  });
get_list [s] u r =
  (case Loop.get s u r of {
    Nothing -> Nothing;
    Just d -> Just [d];
  });
get_list [] u r = Nothing;

get_loop ::
  forall a b.
    (Eq a,
      Eq b) => [(Term.Term a b)] ->
                 Term.Ctxt a b ->
                   Term.Sub a b ->
                     [(Term.Term a b, Term.Term a b)] ->
                       Maybe [(Term.Term a b,
                                (Term.Ctxt a b,
                                  (Term.Sub a b,
                                    (Term.Term a b, Term.Term a b))))];
get_loop (t : ts) d tau r =
  Loop.get_list (t : ts) (Term.ctxt_apply d (Term.sub_apply t tau)) r;
get_loop [] uu uv uw = Nothing;

rewrites_to ::
  forall a b.
    (Eq a,
      Eq b) => [(Term.Term a b,
                  (Term.Ctxt a b,
                    (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                 -> Term.Term a b -> [(Term.Term a b, Term.Term a b)] -> Bool;
rewrites_to ((s, (c, (sigma, rule))) : (t, (d, (tau, rule'))) : xs) u r =
  Loop.rewrites (s, t) c sigma rule r &&
    Loop.rewrites_to ((t, (d, (tau, rule'))) : xs) u r;
rewrites_to [(s, (c, (sigma, rule)))] t r = Loop.rewrites (s, t) c sigma rule r;

check_loop_d ::
  forall a b.
    (Eq a,
      Eq b) => [(Term.Term a b,
                  (Term.Ctxt a b,
                    (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                 -> Term.Ctxt a b ->
                      Term.Sub a b -> [(Term.Term a b, Term.Term a b)] -> Bool;
check_loop_d (v : va) d tau r =
  Loop.rewrites_to (v : va)
    (Term.ctxt_apply d (Term.sub_apply (fst (List.hd (v : va))) tau)) r;
check_loop_d [] uu uv uw = False;

check_loop ::
  forall a b.
    (Eq a,
      Eq b) => ([(Term.Term a b)], (Term.Ctxt a b, Term.Sub a b)) ->
                 [(Term.Term a b, Term.Term a b)] -> Bool;
check_loop (ts, (c, sigma)) r =
  (case Loop.get_loop ts c sigma r of {
    Nothing -> False;
    Just ls -> Loop.check_loop_d ls c sigma r;
  });

}
