(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DepGraph imports DpFramework Edg begin 



fun checkDepGraphProc :: "('f :: sshow,'v :: sshow)trsL \<Rightarrow> (bool \<times> ('f,'v)trsL) list \<Rightarrow> ('f, 'v)trsL \<Rightarrow> sstring list monad"
  where "checkDepGraphProc P dps R = (
     checkDecomposition (checkEdgRM R) P dps
   \<leftarrow> (\<lambda> s. ''our estimation (EDG***) could not show that you gave a valid decomposition'' # s)
  )"


lemma checkDepGraphProcSound[simp] :
  assumes check: "isOK (checkDepGraphProc P edps R)" 
  and rec: "\<forall> edp \<in> set edps. fst edp \<longrightarrow> finite_dpp (set (snd edp), set R)" (is ?rec)
  shows "finite_dpp (set P, set R)"
using check
proof auto
  assume scc: "isOK (checkDecomposition (checkEdgRM R) P edps)"
           (is "isOK (checkDecomposition ?g _ _)")
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> ?thesis" 
    from this obtain s t \<sigma> where 
      inP:  "\<forall>i. (s i, t i) \<in> set P" and 
      inBetween: "\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep (set R))^* " and
      minimal: "\<forall>i. (SN (rstep (set R)) (s i \<cdot> \<sigma> i) \<and> SN (rstep (set R)) (t i \<cdot> \<sigma> i))"
      by (unfold finite_dpp_def, auto)
    
    let ?p = "\<lambda> i. (s i, t i)"
    let ?ps = "\<lambda> i. (s i \<cdot> \<sigma> i, t i \<cdot> \<sigma> i)"
    have one: "isPath (graph2 P ?g) ?p"
    proof (unfold edges_def graph2_def isPath_def, simp, clarify, unfold mem_def)
      fix n
      from inBetween have "(t n \<cdot> \<sigma> n, s (Suc n) \<cdot> \<sigma> (Suc n)) \<in> (rstep (set R))^* " ..
      thus "checkEdgRM R (?p n, ?p (Suc n))" by simp
    qed
    
    from inP have two: "pathNodes ?p \<subseteq> set P" unfolding pathNodes_def by auto 
    
    with one two scc
    obtain q C where 
      suffix: "?p =\<dots> q" and 
      mem: "C \<in> set (map snd (filter fst edps))" and 
      ending: "pathNodes q \<subseteq> set C" 
      using checkDecompositionSound by blast
    
    from suffix obtain n where ende: "\<forall> m. (?p (m + n) = q m)" unfolding suffix_def by blast
    
    hence "min_ichain (set C, set R) (shift s n) (shift t n) (shift \<sigma> n)"
      using inP inBetween minimal ending unfolding pathNodes_def by auto
    hence nFin: "\<not> finite_dpp (set C, set R)" unfolding finite_dpp_def by blast

    with mem rec show False by (induct edps, auto simp: mem_iff)
  qed
qed
  
declare checkDepGraphProc.simps[simp del]

end

