(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPair imports Trs DpFramework begin


definition one_step_ctxt_closed :: "('f,'v)trel \<Rightarrow> bool"
where
  "one_step_ctxt_closed r = (\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"

lemma one_step_ctxt_closed_imp_ctxt_closed: "one_step_ctxt_closed r \<Longrightarrow> ctxt_closed r"
proof (unfold one_step_ctxt_closed_def ctxt_closed_def)
  assume one: "(\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  show "CC r \<subseteq> r"
  proof
    fix csct
    assume csct: "csct \<in> CC r"
    thus "csct \<in> r"
    proof (cases csct)
      case (Pair cs ct)
      with csct have "(cs, ct) \<in> CC r" by simp
      thus ?thesis
      proof
	fix s t C 
	assume "cs = C\<langle>s\<rangle>" and "ct = C\<langle>t\<rangle>" and st: "(s,t) \<in> r"
	from st have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> r" 
	proof (induct C arbitrary: s t)
	  case Hole
	  thus ?case by simp
	next
	  case (More f bef C aft)
	  with one show ?case by auto
	qed
	with prems show ?thesis by simp 
      qed
    qed
  qed
qed

lemma CC_monotone : assumes "r \<subseteq> s" shows "CC r \<subseteq> CC s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CC r"
  thus "(u,v) \<in> CC s"
  proof
    fix uu vv C
    assume "u = C\<langle>uu\<rangle>" and "v = C\<langle>vv\<rangle>" and "(uu,vv) \<in> r"
    from prems show ?thesis by (auto simp: ccI)
  qed
qed

lemma CS_monotone : assumes "r \<subseteq> s" shows "CS r \<subseteq> CS s"
using assms
proof auto
  fix u v
  assume "(u,v) \<in> CS r"
  thus "(u,v) \<in> CS s"
  proof
    fix uu vv \<sigma>
    assume "u = uu \<cdot> \<sigma>" and "v = vv \<cdot> \<sigma>" and "(uu,vv) \<in> r"
    from prems show ?thesis by (auto simp: csI)
  qed
qed

lemma R_inclusion : assumes C: "one_step_ctxt_closed Rel" 
  and s: "sub_closed Rel"
  and orient: "R \<subseteq> Rel" 
  shows "rstep R \<subseteq> Rel"
proof -
  from C have D: "ctxt_closed Rel" by (rule one_step_ctxt_closed_imp_ctxt_closed)
  from orient have "CS R \<subseteq> CS Rel" by (rule CS_monotone)
  with s have "CS R \<subseteq> Rel" by (auto simp: sub_closed_def)
  hence "rstep R \<subseteq> CC Rel" by (rule CC_monotone)
  with D show ?thesis by (auto simp: ctxt_closed_def)
qed
      
lemma compat_tr_compat : fixes S assumes "S O NS \<subseteq> S" shows "S O NS^* \<subseteq> S"
proof 
  fix x
  assume seq: "x \<in> S O NS^*"
  show "x \<in> S"
  proof (cases x)
    case (Pair a b)
    with seq have "(a,b) \<in> S O NS^*" by simp
    with `S O NS \<subseteq> S` have "(a,b) \<in> S" by (rule non_strict_into_strict)
    with Pair show ?thesis by simp
  qed
qed



locale redpair =    
  fixes  S :: "('f,'v)trel" ("\<succ>")
    and   NS :: "('f,'v)trel" ("\<succeq>")
  assumes SN: "SN \<succ>"
    and compat: "\<succ> O \<succeq> \<subseteq> \<succ>"
    and wmNS: "one_step_ctxt_closed \<succeq>"
    and subNS: "sub_closed \<succeq>"
    and subS: "sub_closed \<succ>"
begin

lemma mNS: "ctxt_closed \<succeq>" 
  using wmNS by (rule one_step_ctxt_closed_imp_ctxt_closed) 

lemma trCompat : "\<succ> O \<succeq>^* \<subseteq> \<succ>" using compat by (rule compat_tr_compat)



  
lemma redPairSound :
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> NS^* \<and> (?next i) \<in> rP"
  proof
    fix i
    from subNS wmNS rules R_inclusion have "rstep R \<subseteq> NS" by auto
    hence "(rstep R)^* \<subseteq> NS^*" by (rule rtrancl_mono)
    with steps have stepsNS: "(?ts i, ?ss (Suc i)) \<in> NS^*" by blast
    show "(?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> NS^* \<and> (?next i) \<in> rP" 
    proof (cases "?next i \<in> S \<inter> (P \<union> rP)")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> S" using subS by (auto simp: sub_closed_def)
      with stepsNS trCompat show ?thesis by auto 
    next
      case False
      with strict nonStrict allP
      have one: "?next i \<in> NS" and two: "?next i \<in> rP" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> NS" using subNS by (auto simp: sub_closed_def)
      with stepsNS have "(?ts i, ?ts (Suc i)) \<in> NS^*" by auto
      with two show ?thesis by simp
    qed
  qed
  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> NS^* \<union> S" by auto
  from SN have "SN S (?ts 0)" by auto
  from infSeq trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> NS^* - S" by (rule non_strict_ending)
  
  from this obtain j where "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> NS^* - S" ..
  with piece1 have "\<forall> i \<ge> j. ?next i \<in> rP" by blast
  hence "\<forall> i. (s (i+(Suc j)), t (i + (Suc j))) \<in> rP" by auto
  thus "\<exists> j. \<forall> i. (s (i+j), t (i+j)) \<in> rP" ..
qed

lemma redPairProcessorDpp : 
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and finite: "finite_dpp (rP, R)"
  shows "finite_dpp (P,R)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (P,R) s t \<sigma>"
  hence "ichain (P,R) s t \<sigma>" by simp
  with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
  from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
  with finite show False by (unfold finite_dpp_def, blast)
qed
 
lemma redPairProcessorGdpp : 
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> nodes P \<subseteq> NS \<union> S"
  and strict: "nodes P - rP \<subseteq> S"
  and finite: "finite_gdpp (node_restriction P rP, R)" (is "finite_gdpp (?rP, R)")
  shows "finite_gdpp (P,R)"
proof -
  have "\<And> s t \<sigma>. min_ichain (nodes P, R) s t \<sigma> \<Longrightarrow> \<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
  proof - 
    fix s t \<sigma>
    assume min: "min_ichain (nodes P,R) s t \<sigma>"
    hence "ichain (nodes P,R) s t \<sigma>" by simp
    with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
    from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
    with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
    thus "\<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  qed
  from this and finite show ?thesis by (rule dpp_lift_to_gdpp)
qed
end


(* now the implementation starts *)

(*
  fixes checkS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and checkNS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and name :: "sstring" 
    and representation :: "sstring list"
  assumes checkSSound: "isOK(checkS rule) \<Longrightarrow> rule \<in> S" 
    and   checkNSSound: "isOK(checkNS rule)  \<Longrightarrow> rule \<in> NS"
begin
*)


fun checkRedPairProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "checkRedPairProc checkSS checkNSS  name representation P rP r  = 
( let (Sp :: ('f,'v) trsL ,NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) P in
    (
   checkAllM checkNSS r
    \<leftarrow> (\<lambda> s. ''problem when orienting TRS'' # s ) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda> s. ''problem when orienting DPs'' # s )
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the reduction pair processor with the following''
         # (representation @ ([] # s))))
 )"




end